#define CORE_MMR_BASE   0xFFE00000
#define SYS_MMR_BASE    0xFFC00000

#ifndef UPPER
#define UPPER(_x)   ((_x >> 16) & 0xFFFF)
#endif
#ifndef LOWER
#define LOWER(_x)   ( _x        & 0xFFFF)
#endif

#ifndef HI
#define HI(_x)      ( _x >> 16)
#endif
#ifndef LO
#define LO(_x)      ( _x        & 0xFFFF)
#endif


/* Load Immediate 32-bit value/label/address into a data/pointer register */
#define IMM32(reg,val)\
reg##.H=hi(val);\
reg##.L=lo(val)

/* Register Load operations */
#define LD32_REG(reg,val,dreg)\
IMM32(dreg,val);\
reg=dreg

#define LD32_REG_STACK(reg,val)\
[--SP]=R0;\
IMM32(R0,val);\
reg=R0;\
R0=[SP++]


/****************************
* Context Switch Operations *
****************************/

#define CONTEXT_SAVE_C link 16;[--SP]=(R7:4,P5:3);SP+=-16;[FP+8]=R0;[FP+12]=R1;[FP+16]=R2
#define CONTEXT_RESTORE_C SP+=16;P0=[FP+4];(R7:4,P5:3)=[SP++];unlink;jump (P0)

#define CONTEXT_SAVE_ASM link 0;[--SP]=(R7:1,P5:0);[--SP]=ASTAT;PUSH_DAG_REGS(0);PUSH_LOOP_REGS(0)
#define CONTEXT_RESTORE_ASM POP_LOOP_REGS(0);POP_DAG_REGS(0);ASTAT=[SP++];(R7:1,P5:0)=[SP++];unlink;rts


// link frame_size:
// [SP--] = RETS;
// [SP--] = FP;
// FP = SP;
// SP += -frame_size;
//
// frame_size = 16 (four 4-byte entries)
// [FP+ -4]  = R0       (local variable -> argument #1)
// [FP+ -8]  = R1       (local variable -> argument #2)
// [FP+ -12] = R2       (local variable -> argument #3)
// [FP+ -16] = [FP+20]  (local variable -> argument #4)
//
// unlink:
// SP = FP;
// FP = [SP++];
// RETS = [SP++];

#define CONTEXT_SAVE link 16;[--SP]=(R7:1,P5:0);[--SP]=ASTAT;PUSH_DAG_REGS(0);PUSH_LOOP_REGS(0);P0=[FP+20];[FP+ -16]=P0;[FP+ -12]=R2;[FP+ -8]=R1;[FP+ -4]=R0
#define CONTEXT_RESTORE POP_LOOP_REGS(0);POP_DAG_REGS(0);ASTAT=[SP++];(R7:1,P5:0)=[SP++];R0=[FP+ -16];unlink;rts


/******************************************
* Memory Mapped Register (MMR) Operations *
******************************************/

#define RD32_MMR_STACK(mmr,dreg)\
[--SP] = P5;\
IMM32(P5,mmr);\
dreg = [P5];\
P5 = [SP++]

#define RD32_MMR(mmr,dreg,preg)\
IMM32(preg,mmr);\
dreg = [preg]

#define RD16_MMR_STACK(mmr,dreg)\
[--SP] = P5;\
IMM32(P5,mmr);\
dreg = w[P5](z);\
P5 = [SP++]

#define RD16_MMR(mmr,dreg,preg)\
IMM32(preg,mmr);\
dreg = w[preg](z)


#define WR32_MMR_STACK(mmr,dreg)\
[--SP] = P5;\
IMM32(P5,mmr);\
[P5] = dreg;\
P5 = [SP++]

#define WR32_MMR(mmr,dreg)\
IMM32(P5,mmr);\
[P5] = dreg

#define WR16_MMR_STACK(mmr,dreg)\
[--SP] = P5;\
IMM32(P5,mmr);\
w[P5] = dreg ## .L;\
P5 = [SP++]

#define WR16_MMR(mmr,dreg)\
IMM32(P5,mmr);\
w[P5] = dreg ## .L


#define WR32_MMR_LABEL_STACK(mmr,label)\
[--SP] = (R7:7,P5:5);\
IMM32(P5,mmr);\
IMM32(R7,label);\
[P5] = R7;\
(R7:7,P5:5) = [SP++]

#define WR32_MMR_LABEL(mmr,label,preg,dreg)\
IMM32(preg,mmr);\
IMM32(dreg,label);\
[preg] = dreg

#define WR16_MMR_LABEL_STACK(mmr,label)\
[--SP] = (R7:7,P5:5);\
IMM32(P5,mmr);\
R7 = label(z);\
w[P5] = R7.L;\
(R7:7,P5:5) = [SP++]

#define WR16_MMR_LABEL(mmr,label,preg,dreg)\
IMM32(preg,mmr);\
IMM32(dreg,label);\
w[preg] = dreg


/*****************************
* Stack operations: PUSH/POP *
*****************************/

#define PUSH_LOOP_REGS(nr)\
[--SP] = LC ## nr;\
[--SP] = LT ## nr;\
[--SP] = LB ## nr

#define POP_LOOP_REGS(nr)\
LB ## nr = [SP++];\
LT ## nr = [SP++];\
LC ## nr = [SP++]

#define PUSH_DAG_REGS(nr)\
[--SP] = I ## nr;\
[--SP] = B ## nr;\
[--SP] = M ## nr;\
[--SP] = L ## nr

#define POP_DAG_REGS(nr)\
L ## nr = [SP++];\
M ## nr = [SP++];\
B ## nr = [SP++];\
I ## nr = [SP++]

/* compatibility mode */
#define PUSH_LOOP_REGx(nr) PUSH_LOOP_REGS(nr)
#define POP_LOOP_REGx(nr) POP_LOOP_REGS(nr)
#define PUSH_LOOP_CNTx(nr) PUSH_LOOP_REGS(nr)
#define POP_LOOP_CNTx(nr) POP_LOOP_REGS(nr)
#define PUSH_DAG_REGx(nr) PUSH_DAG_REGS(nr)
#define POP_DAG_REGx(nr) POP_DAG_REGS(nr)


/*************************************
* Register Initialization Operations *
*************************************/

/* General Registers (genreg): [R7:0], [P5:0], SP, FP, [A1:0] */
/* dreg */
#define INIT_D_REGS(val)\
R0=val;\
R1=val;\
R2=val;\
R3=val;\
R4=val;\
R5=val;\
R6=val;\
R7=val

/* preg */
#define INIT_P_REGS(val)\
P0=val;\
P1=val;\
P2=val;\
P3=val;\
P4=val;\
P5=val

/* accreg */
#define INIT_ACC_REGS(val)\
A0=val;\
A1=val


/* DAG Registers (dagreg): [B3:0], [I3:0], [M3:0], [L3:0] */
#define INIT_I_REGS(val)\
I0=val;\
I1=val;\
I2=val;\
I3=val

#define INIT_M_REGS(val)\
M0=val;\
M1=val;\
M2=val;\
M3=val

#define INIT_L_REGS(val)\
L0=val;\
L1=val;\
L2=val;\
L3=val

#define INIT_B_REGS(val)\
B0=val;\
B1=val;\
B2=val;\
B3=val

#define INIT_DAG_REGS0(val)\
I0=val;\
M0=val;\
L0=val;\
B0=val

#define INIT_DAG_REGS1(val)\
I1=val;\
M1=val;\
L1=val;\
B1=val

#define INIT_DAG_REGS2(val)\
I2=val;\
M2=val;\
L2=val;\
B2=val

#define INIT_DAG_REGS3(val)\
I3=val;\
M3=val;\
L3=val;\
B3=val

#define INIT_DAG_REGS_NR(nr,val)\
INIT_DAG_REGS##nr(val)

#define INIT_DAG_REGS(val)\
INIT_I_REGS(val);\
INIT_M_REGS(val);\
INIT_L_REGS(val);\
INIT_B_REGS(val)


/* System Registers (sysreg) */
/* seqreg */
#define INIT_LOOP_REGS(val)\
[--SP]=P5;\
P5=val;\
LC0=P5;\
LT0=P5;\
LB0=P5;\
LC1=P5;\
LT1=P5;\
LB1=P5;\
P5=[SP++]

#define INIT_CYCLE_REGS(val)\
[--SP]=R7;\
R7=val;\
CYCLES=R7;\
CYCLES2=R7;\
R7=[SP++]

#define INIT_RET_REGS(val)\
[--SP]=R7;\
R7=val;\
RETS=R7;\
RETI=R7;\
RETX=R7;\
RETN=R7;\
R7=[SP++]

#define INIT_SEQ_REGS(val)\
INIT_LOOP_REGS(val);\
INIT_CYCLE_REGS(val);\
INIT_RET_REGS(val)


/***************************
* miscellaneous operations *
***************************/
/* 
 * Set the corresponding bits in a System Register (SR); 
 * All bits set in "mask" will be set in the system register 
 * specified by "sys_reg" bitset_SR(sys_reg, mask), where 
 * sys_reg is the system register and mask are the bits to be set. 
 */
#define bitset_SR(sys_reg, mask)\
[--SP] = (R7:6);\
R7 = sys_reg;\
R6.H = (mask) >> 16;\
R6.L = (mask) & 0xffff;\
R7 = R7 | R6;\
sys_reg = R7;\
CSYNC;\
(R7:6) = [SP++]

/* 
 * Clear the corresponding bits in a System Register (SR); 
 * All bits set in "mask" will be cleared in the SR 
 * specified by "sys_reg" bitclr_SR(sys_reg, mask), where 
 * sys_reg is the SR and mask are the bits to be cleared. 
 */
#define bitclr_SR(sys_reg, mask)\
[--SP] = (R7:6);\
R7 = sys_reg;\
R7 =~ R7;\
R6.L = (mask) & 0xffff;\
R6.H = (mask) >> 16;\
R7 = R7 | R6;\
R7 =~ R7;\
sys_reg = R7;\
CSYNC;\
(R7:6) = [SP++]

/* 
 * Set the corresponding bits in a Memory Mapped Register (MMR); 
 * All bits set in "mask" will be set in the MMR specified by "mmr"
 * bitset_MMR(mmr, mask), where mmr is the MMR and mask are 
 * the bits to be set. 
 */
#define bitset_MMR(mmr, mask)\
[--SP] = (R7:6);\
[--SP] = P5;\
P5.L = mmr & 0xffff;\
P5.H = mmr >> 16;\
R7 = [P5];\
R6.L = (mask) & 0xffff;\
R6.H = (mask) >> 16;\
R7 = R7 | R6;\
[P5] = R7;\
CSYNC;\
P5 = [SP++];\
(R7:6) = [SP++]

/* 
 * Clear the corresponding bits in a Memory Mapped Register (MMR); 
 * All bits set in "mask" will be cleared in the MMR specified by "mmr"
 * bitclr_MMRreg(mmr, mask), where sys_reg is the MMR and mask are 
 * the bits to be cleared. 
 */
#define bitclr_MMR(mmr, mask)\
[--SP] = (R7:6);\
[--SP] = P5;\
P5.L = mmr & 0xffff;\
P5.H = mmr >> 16;\
R7 = [P5];\
R7 =~ R7;\
R6.L = (mask) & 0xffff;\
R6.H = (mask) >> 16;\
R7 = R7 | R6;\
R7 =~ R7;\
[P5] = R7;\
CSYNC;\
P5 = [SP++];\
(R7:6) = [SP++]
