/*****************************************************************************
**                                                                          **
**      Name:   bfstartup                                                   **
**                                                                          **
******************************************************************************

(C) Copyright 2007 - Analog Devices, Inc.  All rights reserved.

File Name:      bfstartup.asm

Date Modified:  01/29/09    v1.0 ap-muc

History:        06/14/06    v0

Tested
Software:       VisualDSP++ 4.0 (Updated June 2006)
                VisualDSP++ 4.5 (Updated Nov. 2006)
                VisualDSP++ 4.5 (Updated Feb. 2007)
                VisualDSP++ 5.0
                VisualDSP++ 5.0.1.0 (Update 1)
                VisualDSP++ 5.0.2.0 (Update 2)
                VisualDSP++ 5.0.3.0 (Update 3)
                VisualDSP++ 5.0.4.0 (Update 4)
                VisualDSP++ 5.0.5.0 (Update 5)

Tested
Hardware:       ADSP-BF518F EZ-Board Rev. 0.1 Silicon Rev. 0.0
                ADSP-BF526 EZ-Board Rev. 0.1 Silicon Rev. 0.0
                ADSP-BF527 EZ-KIT Lite Rev. 1.0 Silicon Rev. 0.0
                ADSP-BF533 EZ-KIT Lite Rev. 1.7 Silicon Rev. 0.4
                ADSP-BF533 EZ-KIT Lite Rev. 2.1 Silicon Rev. 0.5
                ADSP-BF537 EZ-KIT Lite Rev. 1.3 Silicon Rev. 0.2
                ADSP-BF538 EZ-KIT Lite Rev. 1.2 Silicon Rev. 0.3
                ADSP-BF548 EZ-KIT Lite Rev. 1.2 Silicon Rev. 0.0
                ADSP-BF548 EZ-KIT Lite Rev. 1.3 Silicon Rev. 0.0
                ADSP-BF548 EZ-KIT Lite Rev. 2.1 Silicon Rev. 0.1
                ADSP-BF561 EZ-KIT Lite Rev. 1.3 Silicon Rev. 0.3
                ADSP-BF561 EZ-KIT Lite Rev. 2.1 Silicon Rev. 0.5

Supported
Plattforms:     ADSP-BF518F EZ-Board
                ADSP-BF526 EZ-Board
                ADSP-BF527 EZ-KIT Lite
                ADSP-BF533 EZ-KIT Lite Rev 1.6 and below (modifications required, only 32MB SDRAM installed)
                ADSP-BF533 EZ-KIT Lite Rev 1.7 and above
                ADSP-BF537 EZ-KIT Lite
                ADSP-BF538 EZ-KIT Lite
                ADSP-BF548 EZ-KIT Lite
                ADSP-BF561 EZ-KIT Lite (Silicon Rev. 0.2 and beyond)

Connections:    ADDS-HPUSB-ICE

Settings:       Default

Purpose:        Generic Startup file for
                ADSP-BF518 EZ-Board
                ADSP-BF526 EZ-Board
                ADSP-BF527 EZ-KIT LITE
                ADSP-BF533 EZ-KIT LITE
                ADSP-BF537 EZ-KIT LITE
                ADSP-BF538 EZ-KIT LITE
                ADSP-BF548 EZ-KIT LITE
                ADSP-BF561 EZ-KIT LITE
                tests.

*****************************************************************************/


/****************************************************************************
 Include Section
*****************************************************************************/

#include "exception.h"
#include "configurations.h"
#include "init_platform.h"
#include "system.h"
#include "macros.h"
#include "main.h"

/*****************************************************************************
 Symbolic constants / definitions
******************************************************************************/

#define SYSCFG_RESET_VAL  0x30

/*****************************************************************************
 Global Variables
******************************************************************************/

.EXTERN ldf_stack_end;

/*****************************************************************************
 Functions
******************************************************************************/

.SECTION L1_startup;

/*****************************************************************************
 Prototypes
******************************************************************************/

.GLOBAL __reset;

/****************************************************************************
 Function:      __reset
 
 Description:   Initialize System Configuration Register
*****************************************************************************/

__reset:

/****************************************************************************
 Description:   Set up supervisor stack
****************************************************************************/

IMM32(SP,ldf_stack_end);
FP = SP;

/****************************************************************************
 Description:   Initialize System Configuration Register
****************************************************************************/

R7 = SYSCFG_RESET_VAL;
SYSCFG = R7;
IMM32(P5,SYS_MMR_BASE);

/*****************************************************************************
 Pre-Init Section
******************************************************************************/

/************************************************
* System Interrupt Wakeup Registers (SIC_IWRx)  *
* Required for PLL_CTL and VR_CTL configuration *
* Processor idle and wake up after PLL_LOCKCNT  *
* expired.Reset = 0xFFFF FFFF                   *
* disable all other interrupt sources in the    *
* meantime to disable unwanted wake up events   *
************************************************/

#if defined (SET_PLL) || defined (EN_SDRAM)

#if defined (__ADSPBF51x__) || defined (__ADSPBF52x__)
    R7 = IWR1_DISABLE_ALL;
#elif defined (__ADSPBF538__) || defined (__ADSPBF54x__) || defined (__ADSPBF561__)
    R7 = IWR_DISABLE_ALL;
#endif

#if defined __ADSPBF54x__
    [P5 + lo(SIC_IWR2)] = R7;
#endif

#if defined (__ADSPBF51x__) || defined (__ADSPBF52x__) || defined (__ADSPBF538__) || defined (__ADSPBF54x__) || defined (__ADSPBF561__)
    [P5 + lo(SIC_IWR1)] = R7;
#endif

R7 = IRQ_PLL_WAKEUP;
[P5 + lo(SIC_IWR0)] = R7;

#endif /* defined (SET_PLL) || defined (EN_SDRAM) */


/*****************************************************************************************
 Description:   Dynamic Power Managment
*****************************************************************************************/
DPM:
#if defined SET_PLL
    call __full_on;
#endif


/*****************************************************************************
 Description:   Configure EBIU
******************************************************************************/
EBIU:


/*****************************************************************************
 Description:   Configure and enable asynchronous memory
******************************************************************************/
#if defined EN_ASYNC_MEM

/**********************************************
* Asynchronous Memory Bank Control 0 Register *
**********************************************/

IMM32(R7,EBIU_AMBCTL0_VAL);
[P5 + lo(EBIU_AMBCTL0)] = R7;


/**********************************************
* Asynchronous Memory Bank Control 1 Register *
**********************************************/

IMM32(R7,EBIU_AMBCTL1_VAL);
[P5 + lo(EBIU_AMBCTL1)] = R7;


/**********************************************
* Asynchronous Memory Global Control Register *
***********************************************/

R7.L = EBIU_AMGCTL_VAL;
w[P5 + lo(EBIU_AMGCTL)] = R7;

#endif /* EN_ASYNC_MEM */


/*****************************************************************************
Description:   Configure and enable SDRAM
******************************************************************************/

#if defined EN_SDRAM
    call __sdram_en;
#endif


/*****************************************************************************
 End of Configure EBIU
******************************************************************************/


#if defined (SET_PLL) || defined (EN_SDRAM)

/*********************************************
* Restore System Interrupt Wakeup Registers  *
*********************************************/

#if defined (__ADSPBF51x__) || defined (__ADSPBF52x__)
    R7 = IWR0_ENABLE_ALL;
#else
    R7 = IWR_ENABLE_ALL;
#endif
[P5 + lo(SIC_IWR0)] = R7;

#if defined (__ADSPBF51x__) || defined (__ADSPBF52x__)
    IMM32(R7,IWR1_ENABLE_ALL);
#endif
#if defined (__ADSPBF51x__) || defined (__ADSPBF52x__) || defined (__ADSPBF538__) || defined (__ADSPBF54x__) || defined (__ADSPBF561__)
    [P5 + lo(SIC_IWR1)] = R7;
#endif

#if defined __ADSPBF54x__
    [P5 + lo(SIC_IWR2)] = R7;
#endif

#endif /* defined (SET_PLL) || defined (EN_SDRAM) */


/****************************************************************************
 Description:   Configure Default Service Handlers
*****************************************************************************/

IMM32(P5,EVT2);
IMM32(R7,nmi_handler);
[P5++] = R7;
IMM32(R7,evx_handler);
[P5++] = R7;
/* IVT4 isn't used */
[P5++] = R7;
IMM32(R7,ivhw_handler);
[P5++] = R7;
IMM32(R7,ivtmr_handler);
[P5++] = R7;
IMM32(R7,ivg7_handler);
[P5++] = R7;
IMM32(R7,ivg8_handler);
[P5++] = R7;
IMM32(R7,ivg9_handler);
[P5++] = R7;
IMM32(R7,ivg10_handler);
[P5++] = R7;
IMM32(R7,ivg11_handler);
[P5++] = R7;
IMM32(R7,ivg12_handler);
[P5++] = R7;
IMM32(R7,ivg13_handler);
[P5++] = R7;
IMM32(R7,ivg14_handler);
[P5++] = R7;
IMM32(R7,ivg15_handler);
[P5++] = R7;

/****************************************************************************
 Description:   Configure and Enable CPLB / Cache
*****************************************************************************/

#if defined INSTR_CACHE_EN
//    call __icplb_init;
#endif
#if defined DATA_CACHE_EN
//    call __dcplb_init;
#endif


/****************************************************************************
 Description:   Clean up everything 
*****************************************************************************/

/* General Registers (genreg): [R7:0], [P5:0], SP, FP, [A1:0] */
/* dreg */
INIT_D_REGS(0);
/* preg */
INIT_P_REGS(0);
/* accreg */
INIT_ACC_REGS(0);
/* DAG Registers (dagreg): [B3:0], [I3:0], [M3:0], [L3:0] */
INIT_DAG_REGS(0);
/* System Registers (sysreg) */
/* seqreg */
INIT_SEQ_REGS(0);
/* astat */
ASTAT = R7;


/****************************************************************************
 Description:   Leave startup code either to User or Supervisor Mode
                jump to main routine
*****************************************************************************/

#ifndef RUN_USER
    /* Enable Interrupt 15 and go to SUPERVISOR mode */
    IMM32(P5,EVT15);
    IMM32(R7,ivg15_handler.supervisor);
    [P5] = R7;

    R7 = ( EVT_IVHW | EVT_IVG15 ) (z);
    sti R7;

    IMM32(P5,WaitForIvg15);
    RETI = P5;                  /* Set Reset Return Address */

    R7 = 0;
    raise 15;
    /* Return from RESET (IVG2) to SUPERVISOR (IVG15) mode */
    rti;

    WaitForIvg15: jump WaitForIvg15; /* Wait here till IVG15 interrupt is requested */
    nop;nop;nop;nop;nop;

    ivg15_handler.supervisor:
    [--SP] = RETI; /* RETI must be pushed to enable interrupts. */
    IMM32(P5,_main);
    jump (P5);
    ivg15_handler.supervisor.end:
#else
    R7 = EVT_IVHW (z);
    sti R7;

    IMM32(P5,_main);
    RETI = P5;

    R7 = 0;
    P5 = R7;
    /* Return from RESET (IVG2) to USER mode (no IVG) */
    rti;
#endif


/****************************************************************************
 Description:  Configure default Interrupt / Exception Handler
*****************************************************************************/

nmi_handler:
EMUEXCPT;
jump 0;
nmi_handler.end:

evx_handler:            /* Exception Handler 3 */

[--SP] = (R7:5,P5:5);
[--SP] = ASTAT;

IMM32(P5,CORE_MMR_BASE);

R7 = SEQSTAT;
R6 = EX_BITS;
R5 = R7 & R6;

#if defined DATA_CACHE_EN
/**********************
* Was it a DCPLB miss *
**********************/
R6 = EX_SYS_DCPLBMISS;
CC = R5 == R6;
IMM32(R7,__dcplb_handler);
if CC jump evx_handler.leave (BP);
#endif

#if defined INSTR_CACHE_EN
/**********************
* Was it a ICPLB miss *
**********************/
R6 = EX_SYS_CCPLBMISS;
CC = R5 == R6;
IMM32(R7,__icplb_handler);
if CC jump evx_handler.leave (BP);
#endif

#if defined (DATA_CACHE_EN) && !defined (DATA_CACHE_WT)
/*************************************************
* Was it a Data access CPLB protection violation *
*************************************************/
R6 = EX_SYS_DCPLBPROT;
CC = R5 == R6;
IMM32(R7,__dcplb_handler);
if CC jump evx_handler.leave (BP);
#endif

#if defined NULLPTR_PROTECTION
/*****************************************
* Was it a Data access multiple CPLB hit *
*****************************************/
R6 = EX_SYS_DCPLBMHIT;
CC = R5 == R6;
/* no handler installed */
if CC jump evx_handler.multiple_cplb_hit (BP);
#endif

/********************************************
* I don't know how to handle this exception *
********************************************/
evx_handler.unknown_exception:
ASTAT = [SP++];
(R7:5,P5:5) = [SP++];
EMUEXCPT;
jump 0;


#if defined NULLPTR_PROTECTION
    evx_handler.multiple_cplb_hit:
    ASTAT = [SP++];
    (R7:5,P5:5) = [SP++];
    /* no handler installed */
    EMUEXCPT;
    jump 0;
#endif


evx_handler.leave:
[P5 + lo(EVT14)] = R7;
R7 = [P5 + lo(IMASK)];
bitset (R7,bitpos(EVT_IVG14));
[P5 + lo(IMASK)] = R7;
raise (bitpos(EVT_IVG14));

ASTAT = [SP++];
(R7:5,P5:5) = [SP++];
/* jump to handler */
rtx;
WaitForIvg14: jump WaitForIvg14; /* Wait here till IVG14 interrupt is requested */
nop;nop;nop;nop;nop;
evx_handler.end:

ivhw_handler:
EMUEXCPT;
jump 0;
ivhw_handler.end:

ivtmr_handler:
EMUEXCPT;
jump 0;
ivtmr_handler.end:

ivg7_handler:
EMUEXCPT;
jump 0;
ivg7_handler.end:

ivg8_handler:
EMUEXCPT;
jump 0;
ivg8_handler.end:

ivg9_handler:
EMUEXCPT;
jump 0;
ivg9_handler.end:

ivg10_handler:
EMUEXCPT;
jump 0;
ivg10_handler.end:

ivg11_handler:
EMUEXCPT;
jump 0;
ivg11_handler.end:

ivg12_handler:
EMUEXCPT;
jump 0;
ivg12_handler.end:

ivg13_handler:
EMUEXCPT;
jump 0;
ivg13_handler.end:

ivg14_handler:
EMUEXCPT;
jump 0;
ivg14_handler.end:

ivg15_handler:
EMUEXCPT;
jump 0;
ivg15_handler.end:


__reset.end:


/*****************************************************************************
 EOF
******************************************************************************/
